#if !defined(NDEBUG)
//#	define BOOST_SPIRIT_X3_DEBUG
#endif

#include "parser.hpp"

#include <boost/spirit/home/x3.hpp>
#include <boost/fusion/adapted/std_pair.hpp>
#if !defined(NDEBUG)
#	include <iostream>
#endif

namespace boost::spirit::x3::traits {
template <typename Char, typename Trait>
struct is_range<std::basic_string_view<Char, Trait>> : boost::mpl::true_ {};
} //namespace boost::spirit::x3::traits

namespace duck {
namespace x3 = boost::spirit::x3;

#if defined(BOOST_SPIRIT_X3_DEBUG)
std::ostream& operator<<(std::ostream& os, const std::vector<TorrentValue>& ot) {
	os << "<list-" << ot.size() << '>';
	return os;
}

std::ostream& operator<<(std::ostream& os, const std::map<TorrentStringType, TorrentValue>& ot) {
	os << "<map-" << ot.size() << '>';
	return os;
}
#endif

namespace workaround {
template <typename Subject>
struct raw_directive : x3::raw_directive<Subject> {
	using x3::raw_directive<Subject>::raw_directive;

	template <typename Iterator, typename Context, typename RContext, typename Attribute>
	bool parse(
		Iterator& first,
		Iterator const& last,
		Context const& context,
		RContext& rcontext,
		Attribute& attr
	) const {
		x3::skip_over(first, last, context);
		Iterator saved = first;
		if (this->subject.parse(first, last, context, rcontext, x3::unused)) {
			attr = { saved, typename Attribute::size_type(first - saved) };
			return true;
		}
		return false;
	}
};

struct raw_gen {
	template <typename Subject>
	raw_directive<typename x3::extension::as_parser<Subject>::value_type>
	operator[](Subject subject) const {
		return { x3::as_parser(std::move(subject)) };
	}
};

auto const raw = raw_gen{};

template <typename T>
struct as_type {
	template <typename E>
	constexpr auto operator[](E e) const { return x3::rule<struct _, T> {} = e; }
};

template <typename T>
static inline constexpr as_type<T> as;
} //namespace workaround

namespace {
namespace parser {

using x3::lit;
using x3::char_;
using x3::uint_;
using x3::string;
using x3::eps;
using x3::omit;
using workaround::raw;
using workaround::as;
//using x3::ascii::digit;

class len_tag; //see https://github.com/boostorg/spirit/issues/680

const x3::rule<class TorrentID, std::vector<TorrentValue>> torrent = "torrent";
const x3::rule<class TorrentValueID, TorrentValue> torrent_value = "torrent_value";
const x3::rule<class TorrentIntID, int> torrent_int = "torrent_int";
const x3::rule<class TorrentStringID, TorrentStringType, true> torrent_string = "torrent_string";
const x3::rule<class TorrentListID, std::vector<TorrentValue>> torrent_list = "torrent_list";
const x3::rule<class TorrentDictID, std::map<TorrentStringType, TorrentValue>> torrent_dict = "torrent_dict";

auto torrent_value_def = torrent_int | torrent_string | torrent_list | torrent_dict;

//auto torrent_int_def = lit('i') >> (string("0") | (char_("1-9") >> *digit)) >> lit('e');
auto torrent_int_def = lit('i') >> x3::long_long >> lit('e');

auto torrent_string_def = x3::with<len_tag>(0u)[
	omit[uint_[([](auto& ctx) { x3::get<len_tag>(ctx) = x3::_attr(ctx); })] >> ':'] >> raw[*(
		eps[([](auto& ctx) { x3::_pass(ctx) = x3::get<len_tag>(ctx)-- > 0; })] >> as<TorrentStringType>[char_]
	)]
];
auto torrent_list_def = lit('l') >> *torrent_value >> 'e';
auto torrent_dict_def = lit('d') >> *(torrent_string >> torrent_value) >> lit('e');
auto torrent_def = +torrent_value;

BOOST_SPIRIT_DEFINE(torrent_int, torrent_string, torrent_list, torrent_dict, torrent_value, torrent);
} //namespace parser
} //unnamed namespace

std::vector<TorrentValue> parse_torrent (std::string_view binary_data) {
	using boost::spirit::x3::parse;

	const char* begin = binary_data.data();
	const char* const end = binary_data.data() + binary_data.size();
	std::vector<TorrentValue> retval;

	const bool r = parse(begin, end, parser::torrent, retval);

#if !defined(NDEBUG)
	std::cout << "done, r=" << r << " (" << begin-binary_data.data() << ")\n";
#endif

	return retval;
}
} //namespace duck
